"""BaseCache for foundation of app-specific caching strategy."""

from calendar import timegm
from datetime import date, datetime, timedelta
from pytz import utc
import json

from django.conf import settings
from django.db.models.loading import get_model

from .models import PkOnlyModel, PkOnlyQueryset


class BaseCache(object):

    """Base instance cache.

    To make the cache useful, create a derived class with methods for
    your Django models.  See drf_cached_instances/tests/test_user_example.py
    for an example.
    """

    default_version = 'default'
    versions = ['default']

    def __init__(self):
        """Initialize BaseCache."""
        self._cache = None
        assert self.default_version in self.versions

    @property
    def cache(self):
        """Get the Django cache interface.

        This allows disabling the cache with
        settings.USE_DRF_INSTANCE_CACE=False.  It also delays import so that
        Django Debug Toolbar will record cache requests.
        """
        if not self._cache:
            use_cache = getattr(settings, 'USE_DRF_INSTANCE_CACHE', True)
            if use_cache:
                from django.core.cache import cache
                self._cache = cache
        return self._cache

    def key_for(self, version, model_name, obj_pk):
        """Get the cache key for the cached instance."""
        return 'drfc_{0}_{1}_{2}'.format(version, model_name, obj_pk)

    def model_function(self, model_name, version, func_name):
        """Return the model-specific caching function."""
        assert func_name in ('serializer', 'loader', 'invalidator')
        name = "%s_%s_%s" % (model_name.lower(), version, func_name)
        return getattr(self, name)

    def field_function(self, type_code, func_name):
        """Return the field function."""
        assert func_name in ('to_json', 'from_json')
        name = "field_%s_%s" % (type_code.lower(), func_name)
        return getattr(self, name)

    def field_to_json(self, type_code, key, *args, **kwargs):
        """Convert a field to a JSON-serializable representation."""
        assert ':' not in key
        to_json = self.field_function(type_code, 'to_json')
        key_and_type = "%s:%s" % (key, type_code)
        json_value = to_json(*args, **kwargs)
        return key_and_type, json_value

    def field_from_json(self, key_and_type, json_value):
        """Convert a JSON-serializable representation back to a field."""
        assert ':' in key_and_type
        key, type_code = key_and_type.split(':', 1)
        from_json = self.field_function(type_code, 'from_json')
        value = from_json(json_value)
        return key, value

    def get_instances(self, object_specs, version=None):
        """Get the cached native representation for one or more objects.

        Keyword arguments:
        object_specs - A sequence of triples (model name, pk, obj):
        - model name - the name of the model
        - pk - the primary key of the instance
        - obj - the instance, or None to load it
        version - The cache version to use, or None for default

        To get the 'new object' representation, set pk and obj to None

        Return is a dictionary:
        key - (model name, pk)
        value - (native representation, pk, object or None)
        """
        ret = dict()
        spec_keys = set()
        cache_keys = []
        version = version or self.default_version

        # Construct all the cache keys to fetch
        for model_name, obj_pk, obj in object_specs:
            assert model_name
            assert obj_pk

            # Get cache keys to fetch
            obj_key = self.key_for(version, model_name, obj_pk)
            spec_keys.add((model_name, obj_pk, obj, obj_key))
            cache_keys.append(obj_key)

        # Fetch the cache keys
        if cache_keys and self.cache:
            cache_vals = self.cache.get_many(cache_keys)
        else:
            cache_vals = {}

        # Use cached representations, or recreate
        cache_to_set = {}
        for model_name, obj_pk, obj, obj_key in spec_keys:

            # Load cached objects
            obj_val = cache_vals.get(obj_key)
            obj_native = json.loads(obj_val) if obj_val else None

            # Invalid or not set - load from database
            if not obj_native:
                if not obj:
                    loader = self.model_function(model_name, version, 'loader')
                    obj = loader(obj_pk)
                serializer = self.model_function(
                    model_name, version, 'serializer')
                obj_native = serializer(obj) or {}
                if obj_native:
                    cache_to_set[obj_key] = json.dumps(obj_native)

            # Get fields to convert
            keys = [key for key in obj_native.keys() if ':' in key]
            for key in keys:
                json_value = obj_native.pop(key)
                name, value = self.field_from_json(key, json_value)
                assert name not in obj_native
                obj_native[name] = value

            if obj_native:
                ret[(model_name, obj_pk)] = (obj_native, obj_key, obj)

        # Save any new cached representations
        if cache_to_set and self.cache:
            self.cache.set_many(cache_to_set)

        return ret

    def update_instance(self, model_name, pk, instance=None, version=None):
        """Create or update a cached instance.

        Keyword arguments are:
        model_name - The name of the model
        pk - The primary key of the instance
        instance - The Django model instance, or None to load it
        versions - Version to update, or None for all

        Return is a list of tuples (model name, pk, immediate) that also needs
        to be updated.
        """
        versions = [version] if version else self.versions
        invalid = []
        for version in versions:
            serializer = self.model_function(model_name, version, 'serializer')
            loader = self.model_function(model_name, version, 'loader')
            invalidator = self.model_function(
                model_name, version, 'invalidator')
            if serializer is None and loader is None and invalidator is None:
                continue

            if self.cache is None:
                continue

            # Try to load the instance
            if not instance:
                instance = loader(pk)

            if serializer:
                # Get current value, if in cache
                key = self.key_for(version, model_name, pk)
                current_raw = self.cache.get(key)
                current = json.loads(current_raw) if current_raw else None

                # Get new value
                new = serializer(instance)
                deleted = not instance

                # If cache is invalid, update cache
                invalidate = (current != new) or deleted
                if invalidate:
                    if deleted:
                        self.cache.delete(key)
                    else:
                        self.cache.set(key, json.dumps(new))
            else:
                invalidate = True

            # Invalidate upstream caches
            if instance and invalidate:
                for upstream in invalidator(instance):
                    if isinstance(upstream, str):
                        self.cache.delete(upstream)
                    else:
                        m, i, immediate = upstream
                        if immediate:
                            invalidate_key = self.key_for(version, m, i)
                            self.cache.delete(invalidate_key)
                        invalid.append((m, i, version))
        return invalid

    #
    # Built-in Field converters
    #

    def field_date_from_json(self, date_triple):
        """Convert a date triple to the date."""
        return date(*date_triple) if date_triple else None

    def field_date_to_json(self, day):
        """Convert a date to a date triple."""
        return [day.year, day.month, day.day] if day else None

    def field_datetime_from_json(self, json_val):
        """Convert a UTC timestamp to a UTC datetime."""
        if type(json_val) == int:
            seconds = int(json_val)
            dt = datetime.fromtimestamp(seconds, utc)
        else:
            seconds, microseconds = [int(x) for x in json_val.split('.')]
            dt = datetime.fromtimestamp(seconds, utc)
            dt += timedelta(microseconds=microseconds)
        return dt

    def field_datetime_to_json(self, dt):
        """Convert a datetime to a UTC timestamp w/ microsecond resolution.

        datetimes w/o timezone will be assumed to be in UTC
        """
        ts = timegm(dt.utctimetuple())
        if dt.microsecond:
            return "{0}.{1:0>6d}".format(ts, dt.microsecond)
        else:
            return ts

    def field_pklist_from_json(self, data):
        """Load a PkOnlyQueryset from a JSON dict.

        This uses the same format as cached_queryset_from_json
        """
        model = get_model(data['app'], data['model'])
        return PkOnlyQueryset(self, model, data['pks'])

    def field_pklist_to_json(self, model, pks):
        """Convert a list of primary keys to a JSON dict.

        This uses the same format as cached_queryset_to_json
        """
        app_label = model._meta.app_label
        model_name = model._meta.model_name
        return {
            'app': app_label,
            'model': model_name,
            'pks': list(pks),
        }

    def field_pk_from_json(self, data):
        """Load a PkOnlyModel from a JSON dict."""
        model = get_model(data['app'], data['model'])
        return PkOnlyModel(self, model, data['pk'])

    def field_pk_to_json(self, model, pk):
        """Convert a primary key to a JSON dict."""
        app_label = model._meta.app_label
        model_name = model._meta.model_name
        return {
            'app': app_label,
            'model': model_name,
            'pk': pk,
        }



# -*- coding: utf-8 -*-
"""Model-like classes that work with the instance cache.

These classes are look-alike replacements for django.db.model.Model and
Queryset.  The full interface is not implemented, only enough to use then
in common Django REST Framework use cases.
"""


class PkOnlyModel(object):

    """Emulate a Django model with only the primary key (pk) set.

    This is used to represent related objects.
    """

    def __init__(self, cache, model, pk):
        """Initialize a PkOnlyModel."""
        self.cache = cache
        self.model = model
        self.pk = pk


class PkOnlyQueryset(object):

    """Emulate a Django queryset with only the primary keys (pks) available.

    This is used to represent a group of related objects, which can be
    accessed by iteration (returning PkOnlyModels) or by values_list
    (returning the list of primary keys).
    """

    def __init__(self, cache, model, pks):
        """Initialize PkOnlyQueryset."""
        self.cache = cache
        self.model = model
        self.pks = pks

    def __iter__(self):
        """Return PkOnlyModels for each pk."""
        for pk in self.pks:
            yield PkOnlyModel(self.cache, self.model, pk)

    def all(self):
        """Handle asking for an unfiltered queryset."""
        return self

    def values_list(self, *args, **kwargs):
        """Return the primary keys as a list.

        The only valid call is values_list('pk', flat=True)
        """
        flat = kwargs.pop('flat', False)
        assert flat is True
        assert len(args) == 1
        assert args[0] == self.model._meta.pk.name
        return self.pks


class CachedModel(object):

    """Emulate a Django model, but with data loaded from the cache."""

    def __init__(self, model, data):
        """Initialize a CachedModel."""
        self._model = model
        self._data = data

    def __getattr__(self, name):
        """Return an attribute from the cached data."""
        if name in self._data:
            return self._data[name]
        else:
            raise AttributeError(
                "%r object has no attribute %r" %
                (self.__class__, name))


class CachedQueryset(object):

    """Emulate a Djange queryset, but with data loaded from the cache.

    A real queryset is used to get filtered lists of primary keys, but the
    cache is used instead of the database to get the instance data.
    """

    def __init__(self, cache, queryset, primary_keys=None):
        """Initialize a CachedQueryset."""
        self.cache = cache
        assert queryset is not None
        self.queryset = queryset
        self.model = queryset.model
        self.filter_kwargs = {}
        self._primary_keys = primary_keys

    @property
    def pks(self):
        """Lazy-load the primary keys."""
        if self._primary_keys is None:
            self._primary_keys = list(
                self.queryset.values_list('pk', flat=True))
        return self._primary_keys

    def __iter__(self):
        """Return the cached data as a list."""
        model_name = self.model.__name__
        object_specs = [(model_name, pk, None) for pk in self.pks]
        instances = self.cache.get_instances(object_specs)
        for pk in self.pks:
            model_data = instances.get((model_name, pk), {})[0]
            yield CachedModel(self.model, model_data)

    def all(self):
        """Handle asking for an unfiltered queryset."""
        return self

    def none(self):
        """Handle asking for an empty queryset."""
        return CachedQueryset(self.cache, self.queryset.none(), [])

    def count(self):
        """Return a count of instances."""
        if self._primary_keys is None:
            return self.queryset.count()
        else:
            return len(self.pks)

    def filter(self, **kwargs):
        """Filter the base queryset."""
        assert not self._primary_keys
        self.queryset = self.queryset.filter(**kwargs)
        return self

    def get(self, *args, **kwargs):
        """Return the single item from the filtered queryset."""
        assert not args
        assert list(kwargs.keys()) == ['pk']
        pk = kwargs['pk']
        model_name = self.model.__name__
        object_spec = (model_name, pk, None)
        instances = self.cache.get_instances((object_spec,))
        try:
            model_data = instances[(model_name, pk)][0]
        except KeyError:
            raise self.model.DoesNotExist(
                "No match for %r with args %r, kwargs %r" %
                (self.model, args, kwargs))
        else:
            return CachedModel(self.model, model_data)

    def __getitem__(self, key):
        """Access the queryset by index or range."""
        if self._primary_keys is None:
            pks = self.queryset.values_list('pk', flat=True)[key]
        else:
            pks = self.pks[key]
        return CachedQueryset(self.cache, self.queryset, pks)



"""Mixins to add caching to Django REST Framework viewsets."""
from django.http import Http404
from rest_framework.generics import get_object_or_404 as _get_object_or_404

from .models import CachedQueryset


def get_object_or_404(queryset, *filter_args, **filter_kwargs):
    """Return an object or raise a 404.

    Same as Django's standard shortcut, but make sure to raise 404
    if the filter_kwargs don't match the required types.
    """
    if isinstance(queryset, CachedQueryset):
        try:
            return queryset.get(*filter_args, **filter_kwargs)
        except queryset.model.DoesNotExist:
            raise Http404('No %s matches the given query.' % queryset.model)
    else:
        return _get_object_or_404(queryset, *filter_args, **filter_kwargs)


class CachedViewMixin(object):

    """Mixin to add caching to a DRF viewset.

    A user should either define cache_class or override get_queryset_cache().
    """

    cache_version = 'default'

    def get_queryset(self):
        """Get the queryset for the action.

        If action is read action, return a CachedQueryset
        Otherwise, return a Django queryset
        """
        queryset = super(CachedViewMixin, self).get_queryset()
        if self.action in ('list', 'retrieve'):
            return CachedQueryset(self.get_queryset_cache(), queryset=queryset)
        else:
            return queryset

    def get_queryset_cache(self):
        """Get the cache to use for querysets."""
        return self.cache_class()

    def get_object(self, queryset=None):
        """
        Return the object the view is displaying.

        Same as rest_framework.generics.GenericAPIView, but:
        - Failed assertions instead of deprecations
        """
        # Determine the base queryset to use.
        assert queryset is None, "Passing a queryset is disabled"
        queryset = self.filter_queryset(self.get_queryset())

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        lookup = self.kwargs.get(lookup_url_kwarg, None)
        assert lookup is not None, "Other lookup methods are disabled"
        filter_kwargs = {self.lookup_field: lookup}
        obj = get_object_or_404(queryset, **filter_kwargs)

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj



"""DRF serializers for sample app."""

from django.contrib.auth.models import User
from rest_framework.serializers import (
    DateField, ModelSerializer)

from .models import Question, Choice


class UserSerializer(ModelSerializer):

    """DRF serializer for Users."""

    created = DateField(source='date_joined', read_only=True)

    class Meta:
        model = User
        fields = ('id', 'username', 'created')


class QuestionSerializer(ModelSerializer):

    """DRF serializer for Questions."""

    class Meta:
        model = Question


class ChoiceSerializer(ModelSerializer):

    """DRF serializer for Choices."""

    class Meta:
        model = Choice



"""Cache implementation for sample app."""

from django.contrib.auth.models import User, Group

from drf_cached_instances.cache import BaseCache

from .models import Question, Choice


class SampleCache(BaseCache):

    """Cache for the sample poll cache."""

    def user_default_serializer(self, obj):
        """Convert a User to a cached instance representation."""
        if not obj:
            return None
        self.user_default_add_related_pks(obj)
        return dict((
            ('id', obj.id),
            ('username', obj.username),
            self.field_to_json('DateTime', 'date_joined', obj.date_joined),
            self.field_to_json(
                'PKList', 'votes', model=Choice, pks=obj._votes_pks),
        ))

    def user_default_loader(self, pk):
        """Load a User from the database."""
        try:
            obj = User.objects.get(pk=pk)
        except User.DoesNotExist:
            return None
        else:
            self.user_default_add_related_pks(obj)
            return obj

    def user_default_add_related_pks(self, obj):
        """Add related primary keys to a User instance."""
        if not hasattr(obj, '_votes_pks'):
            obj._votes_pks = list(obj.votes.values_list('pk', flat=True))

    def user_default_invalidator(self, obj):
        """Invalidate cached items when the User changes."""
        return ['drfc_user_count']

    group_default_serializer = None

    def group_default_loader(self, pk):
        """Load a Group from the database."""
        return Group.objects.get(pk=pk)

    def group_default_invalidator(self, obj):
        """Invalidated cached items when the Group changes."""
        user_pks = User.objects.values_list('pk', flat=True)
        return [('User', pk, False) for pk in user_pks]

    bar_default_serializer = None
    bar_default_loader = None
    bar_default_invalidator = None

    def question_default_serializer(self, obj):
        """Convert a Question to a cached instance representation."""
        if not obj:
            return None
        self.question_default_add_related_pks(obj)
        return dict((
            ('id', obj.id),
            ('question_text', obj.question_text),
            self.field_to_json('DateTime', 'pub_date', obj.pub_date),
            self.field_to_json(
                'PKList', 'choices', model=Choice, pks=obj._choice_pks),
        ))

    def question_default_loader(self, pk):
        """Load a Question from the database."""
        try:
            obj = Question.objects.get(pk=pk)
        except Question.DoesNotExist:
            return None
        else:
            self.question_default_add_related_pks(obj)
            return obj

    def question_default_add_related_pks(self, obj):
        """Add related primary keys to a Question instance."""
        if not hasattr(obj, '_choice_pks'):
            obj._choice_pks = list(obj.choices.values_list('pk', flat=True))

    def question_default_invalidator(self, obj):
        """Invalidated cached items when the Question changes."""
        return []

    def choice_default_serializer(self, obj):
        """Convert a Choice to a cached instance representation."""
        if not obj:
            return None
        self.choice_default_add_related_pks(obj)
        return dict((
            ('id', obj.id),
            ('choice_text', obj.choice_text),
            self.field_to_json(
                'PK', 'question', model=Question, pk=obj.question_id),
            self.field_to_json(
                'PKList', 'voters', model=User, pks=obj._voter_pks)
        ))

    def choice_default_loader(self, pk):
        """Load a Choice from the database."""
        try:
            obj = Choice.objects.get(pk=pk)
        except Choice.DoesNotExist:
            return None
        else:
            self.choice_default_add_related_pks(obj)
            return obj

    def choice_default_add_related_pks(self, obj):
        """Add related primary keys to a Choice instance."""
        if not hasattr(obj, '_voter_pks'):
            obj._voter_pks = obj.voters.values_list('pk', flat=True)

    def choice_default_invalidator(self, obj):
        """Invalidated cached items when the Choice changes."""
        invalid = [('Question', obj.question_id, True)]
        for pk in obj.voters.values_list('pk', flat=True):
            invalid.append(('User', pk, False))
        return invalid



"""Asynchronous tasks for sample_poll_app."""
from celery import shared_task

from .cache import SampleCache


@shared_task(ignore_result=True)
def update_cache_for_instance(
        model_name, instance_pk, instance=None, version=None):
    """Update the cache for an instance, with cascading updates."""
    cache = SampleCache()
    invalid = cache.update_instance(model_name, instance_pk, instance, version)
    for invalid_name, invalid_pk, invalid_version in invalid:
        update_cache_for_instance.delay(
            invalid_name, invalid_pk, version=invalid_version)



"""URL routes for the sample app."""
from django.conf.urls import include, patterns, url
from django.views.generic import TemplateView

from rest_framework.routers import DefaultRouter

from .viewsets import ChoiceViewSet, QuestionViewSet, UserViewSet

router = DefaultRouter()
router.register(r'users', UserViewSet)
router.register(r'questions', QuestionViewSet)
router.register(r'choices', ChoiceViewSet)

urlpatterns = patterns(
    '',
    url(r'^$', TemplateView.as_view(
        template_name='sample_poll_app/home.html'), name='home'),
    url(r'^api-auth/', include(
        'rest_framework.urls', namespace='rest_framework')),
    url(r'^api/', include(router.urls))
)



"""Django REST Framework viewsets for sample app."""
from django.contrib.auth.models import User
from drf_cached_instances.mixins import CachedViewMixin
from rest_framework.viewsets import ModelViewSet as BaseModelViewSet

from .cache import SampleCache
from .models import Choice, Question
from .serializers import ChoiceSerializer, QuestionSerializer, UserSerializer


class ModelViewSet(CachedViewMixin, BaseModelViewSet):

    """ModelViewSet that uses CachedViewMixin."""

    cache_class = SampleCache


class UserViewSet(ModelViewSet):

    """API endpoint that allows users to be viewed or edited."""

    queryset = User.objects.all()
    serializer_class = UserSerializer


class QuestionViewSet(ModelViewSet):

    """API endpoint that allows questions to be viewed or edited."""

    queryset = Question.objects.all()
    serializer_class = QuestionSerializer


class ChoiceViewSet(ModelViewSet):

    """API endpoint that allows choices to be viewed or edited."""

    queryset = Choice.objects.all()
    serializer_class = ChoiceSerializer



"""Tests for drf_cached_instances/models.py. """
from django.contrib.auth.models import User
from django.test import TestCase

from drf_cached_instances.models import (
    CachedModel, CachedQueryset, PkOnlyModel, PkOnlyQueryset)

from sample_poll_app.cache import SampleCache


class TestCachedModel(TestCase):

    """Tests for CachedModel."""

    def test_has_data(self):
        """Data can be accessed by attribute."""
        cm = CachedModel(User, {'username': 'frank'})
        self.assertEqual('frank', cm.username)

    def test_does_not_have_data(self):
        """Accessing missing attributes is AttributeError, not KeyError."""
        cm = CachedModel(User, {'username': 'frank'})
        self.assertRaises(AttributeError, getattr, cm, 'email')


class TestCachedQueryset(TestCase):

    """Tests for TestCachedQueryset."""

    def setUp(self):
        """Shared objects for testing."""
        self.cache = SampleCache()
        self.cache.cache.clear()

    def create_users(self, number):
        """Create a set of test users."""
        for x in range(number):
            User.objects.create(username='user%d' % x)

    def test_get_existing_instance(self):
        """A cached instance can be retrieved by get(pk)."""
        user = User.objects.create(username='frank')
        cq = CachedQueryset(self.cache, User.objects.all())
        cached_user = cq.get(pk=user.pk)
        self.assertEqual('frank', cached_user.username)

    def test_get_nonexisting_instance(self):
        """Attempting to get a missing instance raises DoesNotExist."""
        self.assertFalse(User.objects.filter(pk=666).exists())
        cq = CachedQueryset(self.cache, User.objects.all())
        self.assertRaises(User.DoesNotExist, cq.get, pk=666)

    def test_all(self):
        """Filtering by all() returns the CachedQueryset."""
        cq = CachedQueryset(self.cache, User.objects.all())
        self.assertEqual(cq, cq.all())

    def test_none(self):
        """An empty queryset has no items."""
        cq = CachedQueryset(self.cache, User.objects.all())
        cq_none = cq.none()
        self.assertEqual([], cq_none.pks)

    def test_pks_by_queryset(self):
        """Acessing primary keys populates the pk cache from the database."""
        self.create_users(10)
        cq = CachedQueryset(self.cache, User.objects.all())
        with self.assertNumQueries(1):
            pks = cq.pks
        self.assertEqual(10, len(pks))

    def test_pks_by_pks(self):
        """Accessing cached primary keys does not query the database."""
        self.create_users(10)
        cq = CachedQueryset(self.cache, User.objects.all(), [1, 2, 3])
        with self.assertNumQueries(0):
            pks = cq.pks
        self.assertEqual(3, len(pks))
        self.assertEqual([1, 2, 3], pks)

    def test_iteration_of_populated_queryset(self):
        """Iterating through a queryset returns CachedModels."""
        self.create_users(10)
        user_pks = list(User.objects.values_list('pk', flat=True))
        cq = CachedQueryset(self.cache, User.objects.order_by('pk'))
        for pk, cm in zip(user_pks, cq):
            self.assertIsInstance(cm, CachedModel)
            self.assertEqual(pk, cm.id)

    def test_iteration_of_empty_queryset(self):
        """Iterating through a queryset returns CachedModels."""
        cq = CachedQueryset(self.cache, User.objects.order_by('pk'))
        self.assertFalse(list(cq))

    def test_count_by_queryset(self):
        """Getting the count with an empty cache is a database operation."""
        self.create_users(10)
        cq = CachedQueryset(self.cache, User.objects.all())
        with self.assertNumQueries(1):
            self.assertEqual(10, cq.count())
        self.assertIsNone(cq._primary_keys)

    def test_count_by_pks(self):
        """Getting the count with a full cache does not query the database."""
        self.create_users(10)
        cq = CachedQueryset(self.cache, User.objects.all(), range(5))
        with self.assertNumQueries(0):
            self.assertEqual(5, cq.count())

    def test_filter(self):
        """A queryset without pks retrived can be filtered by pk."""
        self.create_users(5)
        user = User.objects.latest('id')
        cq = CachedQueryset(self.cache, User.objects.all())
        users = list(cq.filter(id=user.id))
        self.assertEqual(1, len(users))
        self.assertEqual(user.id, users[0].id)

    def test_get_slice_by_queryset(self):
        """A queryset slice of postpones database access until usage."""
        self.create_users(10)
        cq = CachedQueryset(self.cache, User.objects.all())
        with self.assertNumQueries(0):
            users = cq[0:5]
        with self.assertNumQueries(1):
            self.assertEqual(5, users.count())

    def test_get_slice_by_pks(self):
        """A queryset slice with a full cache does not query the database."""
        self.create_users(10)
        cq = CachedQueryset(self.cache, User.objects.all(), range(5))
        with self.assertNumQueries(0):
            users = cq[0:5]
        with self.assertNumQueries(0):
            self.assertEqual(5, users.count())


class TestPkOnlvQueryset(TestCase):

    """Tests for PkOnlyQueryset."""

    def setUp(self):
        """Shared objects for testing."""
        self.cache = SampleCache()
        self.cache.cache.clear()

    def test_iter(self):
        """Iterating a PkOnlyQueryset returns PkOnlyModels."""
        pks = [1, 2, 3]
        pkvl = PkOnlyQueryset(self.cache, User, pks)
        for pk, pm in zip(pks, pkvl):
            self.assertIsInstance(pm, PkOnlyModel)
            self.assertEqual(pm.pk, pk)

    def test_iter_empty_list(self):
        """Iterating an PkOnlyQueryset returns an empty list."""
        pkvl = PkOnlyQueryset(self.cache, User, [])
        self.assertFalse(list(pkvl))

    def test_filter_all(self):
        """Filtering with .all() returns the PkOnlyQueryset."""
        pkvl = PkOnlyQueryset(self.cache, User, range(5))
        pkall = pkvl.all()
        self.assertEqual(pkvl, pkall)

    def test_values_list(self):
        """A limited version of values_list returns pks."""
        pkvl = PkOnlyQueryset(self.cache, User, range(5))
        values_list = pkvl.values_list('id', flat=True)
        self.assertEqual(range(5), values_list)
